from __future__ import annotations
from .pyhornedowl import PyIndexedOntology, open_ontology, get_descendants, get_ancestors

__all__ = ["PyIndexedOntology", "open_ontology", "get_descendants", "get_ancestors"]
